# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This NativeIFC module deals with layers"""

import ifcopenshell
import ifcopenshell.util.element

from . import ifc_tools


def load_layers(obj):
    """Loads all the layers of an IFC file"""

    proj = ifc_tools.get_project(obj)
    ifcfile = ifc_tools.get_ifcfile(obj)
    layers = ifcfile.by_type("IfcPresentationLayerAssignment")
    for layer in layers:
        obj = get_layer(layer, proj)
        populate_layer(obj)


def has_layers(obj):
    """Returns true if the given project has layers"""

    ifcfile = ifc_tools.get_ifcfile(obj)
    layers = ifcfile.by_type("IfcPresentationLayerAssignment")
    if layers:
        return True
    return False


def get_layer(layer, project):
    """Returns (creates if necessary) a layer object in the given project"""

    group = ifc_tools.get_group(project, "IfcLayersGroup")
    if not group:
        return None
    if hasattr(project, "Document"):
        doc = project.Document
    else:
        doc = project
    exobj = ifc_tools.get_object(layer, doc)
    if exobj:
        return exobj
    obj = ifc_tools.add_object(doc, otype="layer")
    ifcfile = ifc_tools.get_ifcfile(project)
    ifc_tools.add_properties(obj, ifcfile, layer)
    group.addObject(obj)
    return obj


def populate_layer(obj):
    """Attaches all the possible objects to this layer"""

    g = []
    element = ifc_tools.get_ifc_element(obj)
    for shape in getattr(element, "AssignedItems", []):
        rep = getattr(shape, "OfProductRepresentation", None)
        for prod in getattr(rep, "ShapeOfProduct", []):
            obj = ifc_tools.get_object(prod)
            if obj:
                g.append(obj)
    obj.Group = g


def add_layers(obj, element=None, ifcfile=None, proj=None):
    """Creates necessary layers for the given object"""

    if not ifcfile:
        ifcfile = ifc_tools.get_ifcfile(obj)
    if not element:
        element = ifc_tools.get_ifc_element(obj, ifcfile)
    if not proj:
        proj = ifc_tools.get_project(obj)
    layers = ifcopenshell.util.element.get_layers(ifcfile, element)
    for layer in layers:
        lay = get_layer(layer, proj)
        if lay and not obj in lay.Group:
            lay.Proxy.addObject(lay, obj)


def add_to_layer(obj, layer):
    """Adds the given object to the given layer"""

    if hasattr(obj, "StepId"):
        obj_element = ifc_tools.get_ifc_element(obj)
    elif hasattr(obj, "id"):
        obj_element = obj
        obj = ifc_tools.get_object(obj_element)
    else:
        return
    if hasattr(layer, "StepId"):
        layer_element = ifc_tools.get_ifc_element(layer)
    elif hasattr(layer, "id"):
        layer_element = layer
        layer = ifc_tools.get_object(layer_element)
    else:
        return
    ifcfile = ifc_tools.get_ifcfile(obj)
    if not ifcfile:
        return
    items = ()
    if layer_element.AssignedItems:
        items = layer_element.AssignedItems
    if not obj_element in items:
        cmd = "attribute.edit_attributes"
        attribs = {"AssignedItems": items + (obj_element,)}
        ifc_tools.api_run(cmd, ifcfile, product=layer_element, attributes=attribs)
    if not obj in layer.Group:
        layer.Proxy.addObject(layer, obj)


def create_layer(name, project):
    """Adds a new layer to the given project"""

    group = ifc_tools.get_group(project, "IfcLayersGroup")
    ifcfile = ifc_tools.get_ifcfile(project)
    try:
        # IfcopenShell 0.8
        layer = ifc_tools.api_run("layer.add_layer", ifcfile, name=name)
    except:
        # IfcopenShell 0.7
        layer = ifc_tools.api_run("layer.add_layer", ifcfile, Name=name)
    return get_layer(layer, project)


def transfer_layer(layer, project):
    """Transfer a non-NativeIFC layer to a project"""

    label = layer.Label
    ifclayer = create_layer(label, project)
    delete = not (ifc_tools.PARAMS.GetBool("KeepAggregated", False))
    # delete the old one if empty and delete param allows
    if delete and not layer.Group:
        layer.Document.removeObject(layer.Name)
    ifclayer.Label = label  # to avoid 001-ing the Label...
    return ifclayer



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2022 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import math

__title__ = "PathLanguage - classes for an internal language/representation for Path"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Functions to extract and convert between Path.Command and Part.Edge and utility functions to reason about them."

CmdMoveStraight = Path.Geom.CmdMoveStraight + Path.Geom.CmdMoveRapid


class Instruction(object):
    """An Instruction is a pure python replacement of Path.Command which also tracks its begin position."""

    def __init__(self, begin, cmd, param=None):
        self.begin = begin
        if isinstance(cmd, Path.Command):
            self.cmd = Path.Name
            self.param = Path.Parameters
        else:
            self.cmd = cmd
            if param is None:
                self.param = {}
            else:
                self.param = param

    def anglesOfTangents(self):
        return (0, 0)

    def setPositionBegin(self, begin):
        self.begin = begin

    def positionBegin(self):
        """positionBegin() ... returns a Vector of the begin position"""
        return self.begin

    def positionEnd(self):
        """positionEnd() ... returns a Vector of the end position"""
        return FreeCAD.Vector(self.x(self.begin.x), self.y(self.begin.y), self.z(self.begin.z))

    def pathLength(self):
        """pathLength() ... returns the length in mm"""
        return 0

    def isMove(self):
        return False

    def isRapid(self):
        return False

    def isPlunge(self):
        """isPlunge() ... return true if this moves is vertical"""
        if self.isMove():
            if (
                Path.Geom.isRoughly(self.begin.x, self.x(self.begin.x))
                and Path.Geom.isRoughly(self.begin.y, self.y(self.begin.y))
                and not Path.Geom.isRoughly(self.begin.z, self.z(self.begin.z))
            ):
                return True
        return False

    def leadsInto(self, instr):
        """leadsInto(instr) ... return true if instr is a continuation of self"""
        return Path.Geom.pointsCoincide(self.positionEnd(), instr.positionBegin())

    def x(self, default=0):
        return self.param.get("X", default)

    def y(self, default=0):
        return self.param.get("Y", default)

    def z(self, default=0):
        return self.param.get("Z", default)

    def i(self, default=0):
        return self.param.get("I", default)

    def j(self, default=0):
        return self.param.get("J", default)

    def k(self, default=0):
        return self.param.get("K", default)

    def xyBegin(self):
        """xyBegin() ... internal convenience function"""
        return FreeCAD.Vector(self.begin.x, self.begin.y, 0)

    def xyEnd(self):
        """xyEnd() ... internal convenience function"""
        return FreeCAD.Vector(self.x(self.begin.x), self.y(self.begin.y), 0)

    def __repr__(self):
        return f"{self.cmd}{self.param}"

    def str(self, digits=2):
        if digits == 0:
            s = [f"{k}: {int(v)}" for k, v in self.param.items()]
        else:
            fmt = f"{{}}: {{:.{digits}}}"
            s = [fmt.format(k, v) for k, v in self.param.items()]
        return f"{self.cmd}{{{', '.join(s)}}}"


class MoveStraight(Instruction):
    def anglesOfTangents(self):
        """anglesOfTangents() ... return a tuple with the tangent angles at begin and end position"""
        begin = self.xyBegin()
        end = self.xyEnd()
        if end == begin:
            return (0, 0)
        a = Path.Geom.getAngle(end - begin)
        return (a, a)

    def isMove(self):
        return True

    def isStraight(self):
        return True

    def isArc(self):
        return False

    def isRapid(self):
        return self.cmd in Path.Geom.CmdMoveRapid

    def pathLength(self):
        return (self.positionEnd() - self.positionBegin()).Length


class MoveArc(Instruction):
    def anglesOfTangents(self):
        """anglesOfTangents() ... return a tuple with the tangent angles at begin and end position"""
        begin = self.xyBegin()
        end = self.xyEnd()
        center = self.xyCenter()
        # calculate angle of the hypotenuse at begin and end
        s0 = Path.Geom.getAngle(begin - center)
        s1 = Path.Geom.getAngle(end - center)
        # the tangents are perpendicular to the hypotenuse with the sign determined by the
        # direction of the arc
        return (
            Path.Geom.normalizeAngle(s0 + self.arcDirection()),
            Path.Geom.normalizeAngle(s1 + self.arcDirection()),
        )

    def isMove(self):
        return True

    def isArc(self):
        return True

    def isStraight(self):
        return False

    def isCW(self):
        return self.arcDirection() < 0

    def isCCW(self):
        return self.arcDirection() > 0

    def arcAngle(self):
        """arcAngle() ... return the angle of the arc opening"""
        begin = self.xyBegin()
        end = self.xyEnd()
        center = self.xyCenter()
        s0 = Path.Geom.getAngle(begin - center)
        s1 = Path.Geom.getAngle(end - center)

        if self.isCW():
            while s0 < s1:
                s0 = s0 + 2 * math.pi
            return s0 - s1

        # CCW
        while s1 < s0:
            s1 = s1 + 2 * math.pi
        return s1 - s0

    def arcRadius(self):
        """arcRadius() ... return the radius"""
        return (self.xyBegin() - self.xyCenter()).Length

    def pathLength(self):
        return self.arcAngle() * self.arcRadius()

    def xyCenter(self):
        return FreeCAD.Vector(self.begin.x + self.i(), self.begin.y + self.j(), 0)


class MoveArcCW(MoveArc):
    def arcDirection(self):
        return -math.pi / 2


class MoveArcCCW(MoveArc):
    def arcDirection(self):
        return math.pi / 2


class Maneuver(object):
    """A series of instructions and moves"""

    def __init__(self, begin=None, instr=None):
        self.instr = instr if instr else []
        self.setPositionBegin(begin if begin else FreeCAD.Vector(0, 0, 0))

    def setPositionBegin(self, begin):
        self.begin = begin
        for i in self.instr:
            i.setPositionBegin(begin)
            begin = i.positionEnd()

    def positionBegin(self):
        return self.begin

    def getMoves(self):
        return [instr for instr in self.instr if instr.isMove()]

    def addInstruction(self, instr):
        self.instr.append(instr)

    def addInstructions(self, coll):
        self.instr.extend(coll)

    def toPath(self):
        return Path.Path([instruction_to_command(instr) for instr in self.instr])

    def __repr__(self):
        if self.instr:
            return "\n".join([str(i) for i in self.instr])
        return ""

    @classmethod
    def InstructionFromCommand(cls, cmd, begin=None):
        if not begin:
            begin = FreeCAD.Vector(0, 0, 0)

        if cmd.Name in CmdMoveStraight:
            return MoveStraight(begin, cmd.Name, cmd.Parameters)
        if cmd.Name in Path.Geom.CmdMoveCW:
            return MoveArcCW(begin, cmd.Name, cmd.Parameters)
        if cmd.Name in Path.Geom.CmdMoveCCW:
            return MoveArcCCW(begin, cmd.Name, cmd.Parameters)
        return Instruction(begin, cmd.Name, cmd.Parameters)

    @classmethod
    def FromPath(cls, path, begin=None):
        maneuver = Maneuver(begin)
        instr = []
        begin = maneuver.positionBegin()
        for cmd in path.Commands:
            i = cls.InstructionFromCommand(cmd, begin)
            instr.append(i)
            begin = i.positionEnd()
        maneuver.instr = instr
        return maneuver

    @classmethod
    def FromGCode(cls, gcode, begin=None):
        return cls.FromPath(Path.Path(gcode), begin)


def instruction_to_command(instr):
    return Path.Command(instr.cmd, instr.param)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
from __future__ import annotations
import abc
from abc import ABC
from typing import Mapping, List, Optional, Type, TYPE_CHECKING
from .uri import AssetUri

if TYPE_CHECKING:
    from .serializer import AssetSerializer


class Asset(ABC):
    asset_type: str

    def __init__(self, *args, **kwargs):
        if not hasattr(self, "asset_type"):
            raise ValueError("Asset subclasses must define 'asset_type'.")

    @property
    def label(self) -> str:
        return self.__class__.__name__

    @abc.abstractmethod
    def get_id(self) -> str:
        """Returns the unique ID of an asset object."""
        pass

    def get_uri(self) -> AssetUri:
        return AssetUri.build(asset_type=self.asset_type, asset_id=self.get_id())

    @classmethod
    def resolve_name(cls, identifier: str) -> AssetUri:
        """
        Resolves an identifier (id, name, or URI) to an AssetUri object.
        """
        # 1. If the input is a url string, return the Uri object for it.
        if AssetUri.is_uri(identifier):
            return AssetUri(identifier)

        # 2. Construct the Uri using Uri.build() and return it
        return AssetUri.build(
            asset_type=cls.asset_type,
            asset_id=identifier,
        )

    @classmethod
    def get_uri_from_id(cls, asset_id):
        return AssetUri.build(cls.asset_type, asset_id=asset_id)

    @classmethod
    def extract_dependencies(cls, data: bytes, serializer: Type[AssetSerializer]) -> List[AssetUri]:
        """Extracts URIs of dependencies from serialized data."""
        return serializer.extract_dependencies(data)

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
        serializer: Type[AssetSerializer],
    ) -> Asset:
        """
        Creates an object from serialized data and resolved dependencies.
        If dependencies is None, it indicates a shallow load where dependencies were not resolved.
        """
        return serializer.deserialize(data, id, dependencies)

    def to_bytes(self, serializer: Type[AssetSerializer]) -> bytes:
        """Serializes an object into bytes."""
        return serializer.serialize(self)



# ***************************************************************************
# *   Copyright (c) 2001 Juergen Riegel <juergen.riegel@web.de>             *
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""FEM module App init script

Gathering all the information to start FreeCAD.
This is the first one of three init scripts.
The third one runs when the gui is up.

The script is executed using exec().
This happens inside srd/Gui/FreeCADGuiInit.py
All imports made there are available here too.
Thus no need to import them here.
But the import code line is used anyway to get flake8 quired.
Since they are cached they will not be imported twice.
"""

__title__ = "FEM module App init script"
__author__ = "Juergen Riegel, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

# imports to get flake8 quired
import sys
import FreeCAD

# needed imports
from femtools.migrate_app import FemMigrateApp


# migrate old FEM App objects
sys.meta_path.append(FemMigrateApp())


# add FEM App unit tests
FreeCAD.__unit_test__ += ["TestFemApp"]


# add import and export file types
FreeCAD.addExportType("FEM mesh Python (*.meshpy)", "feminout.importPyMesh")

FreeCAD.addExportType("FEM mesh TetGen (*.poly)", "feminout.convert2TetGen")

# see FemMesh::read() and FemMesh::write() methods in src/Mod/Fem/App/FemMesh.cpp
FreeCAD.addImportType(
    "FEM mesh formats (*.bdf *.BDF *.dat *.DAT *.inp *.INP *.med *.MED *.unv *.UNV *.vtk *.VTK *.vtu *.VTU *.pvtu *.PVTU *.z88 *.Z88)",
    "Fem",
)
FreeCAD.addExportType("FEM mesh formats (*.dat *.inp *.med *.stl *.unv *.vtk *.vtu *.z88)", "Fem")

FreeCAD.addExportType("FEM mesh Nastran (*.bdf)", "feminout.exportNastranMesh")

FreeCAD.addImportType("FEM result CalculiX (*.frd *.FRD)", "feminout.importCcxFrdResults")

FreeCAD.addImportType("FEM mesh Fenics (*.xml *.XML *.xdmf *.XDMF)", "feminout.importFenicsMesh")
FreeCAD.addExportType("FEM mesh Fenics (*.xml *.xdmf)", "feminout.importFenicsMesh")

FreeCAD.addImportType(
    "FEM mesh YAML/JSON (*.meshyaml *.MESHYAML *.meshjson *.MESHJSON *.yaml *.YAML *.json *.JSON)",
    "feminout.importYamlJsonMesh",
)
FreeCAD.addExportType(
    "FEM mesh YAML/JSON (*.meshyaml *.meshjson *.yaml *.json)",
    "feminout.importYamlJsonMesh",
)

FreeCAD.addImportType("FEM mesh Z88 (*.txt *.TXT)", "feminout.importZ88Mesh")
FreeCAD.addExportType("FEM mesh Z88 (*.txt)", "feminout.importZ88Mesh")

FreeCAD.addImportType("FEM result Z88 displacements (*.txt *.TXT)", "feminout.importZ88O2Results")

if "BUILD_FEM_VTK" in FreeCAD.__cmake__:
    FreeCAD.addImportType(
        "FEM result VTK (*.vtk *.VTK *.vtu *.VTU *.pvtu *.PVTU *.vtm *.VTM, *.pvd)",
        "feminout.importVTKResults",
    )
    FreeCAD.addExportType(
        "FEM result VTK (*.vtu *.vtp *.vts *.vtr *.vti *.vtm)", "feminout.importVTKResults"
    )



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path.Main.Stock as PathStock

from CAMTests.PathTestUtils import PathTestBase


class FakeJobProxy:
    def baseObject(self, obj):
        return obj.Base


R = 223.606798 / 2


class TestPathStock(PathTestBase):
    def setUp(self):
        self.doc = FreeCAD.newDocument("TestPathStock")
        self.base = self.doc.addObject("Part::Box", "Box")
        self.base.Length = 100
        self.base.Width = 200
        self.base.Height = 300
        self.job = self.doc.addObject("App::FeaturePython", "Job")
        self.job.addProperty("App::PropertyLink", "Model")
        model = FreeCAD.ActiveDocument.addObject("App::DocumentObjectGroup", "Model")
        model.addObject(self.base)
        self.job.Model = model
        self.job.Proxy = FakeJobProxy()

    def tearDown(self):
        FreeCAD.closeDocument("TestPathStock")

    def test00(self):
        """Test CreateBox"""

        stock = PathStock.CreateBox(self.job)
        self.assertTrue(hasattr(stock, "Length"))
        self.assertTrue(hasattr(stock, "Width"))
        self.assertTrue(hasattr(stock, "Height"))
        self.assertEqual(100, stock.Length)
        self.assertEqual(200, stock.Width)
        self.assertEqual(300, stock.Height)

        extent = FreeCAD.Vector(17, 13, 77)
        stock = PathStock.CreateBox(self.job, extent)
        self.assertEqual(17, stock.Length)
        self.assertEqual(13, stock.Width)
        self.assertEqual(77, stock.Height)

        placement = FreeCAD.Placement(FreeCAD.Vector(-3, 88, 4), FreeCAD.Vector(0, 0, 1), 180)
        stock = PathStock.CreateBox(self.job, extent, placement)
        self.assertEqual(17, stock.Length)
        self.assertEqual(13, stock.Width)
        self.assertEqual(77, stock.Height)
        self.assertPlacement(placement, stock.Placement)

    def test01(self):
        """Test CreateCylinder"""

        stock = PathStock.CreateCylinder(self.job)
        self.assertTrue(hasattr(stock, "Radius"))
        self.assertTrue(hasattr(stock, "Height"))
        self.assertRoughly(R, stock.Radius.Value)
        self.assertEqual(300, stock.Height)

        stock = PathStock.CreateCylinder(self.job, 37, 24)
        self.assertEqual(37, stock.Radius)
        self.assertEqual(24, stock.Height)

        placement = FreeCAD.Placement(FreeCAD.Vector(3, 8, -4), FreeCAD.Vector(0, 0, 1), -90)
        stock = PathStock.CreateCylinder(self.job, 1, 88, placement)
        self.assertEqual(1, stock.Radius)
        self.assertEqual(88, stock.Height)
        self.assertPlacement(placement, stock.Placement)

    def test10(self):
        """Verify FromTemplate box creation."""

        extent = FreeCAD.Vector(17, 13, 77)
        placement = FreeCAD.Placement(FreeCAD.Vector(3, 8, -4), FreeCAD.Vector(0, 0, 1), -90)
        orig = PathStock.CreateBox(self.job, extent, placement)

        # collect full template
        template = PathStock.TemplateAttributes(orig)
        stock = PathStock.CreateFromTemplate(self.job, template)

        self.assertEqual(PathStock.StockType.CreateBox, PathStock.StockType.FromStock(stock))
        self.assertEqual(orig.Length, stock.Length)
        self.assertEqual(orig.Width, stock.Width)
        self.assertEqual(orig.Height, stock.Height)
        self.assertPlacement(orig.Placement, stock.Placement)

        # don't store extent in template
        template = PathStock.TemplateAttributes(orig, False, True)
        stock = PathStock.CreateFromTemplate(self.job, template)

        self.assertEqual(PathStock.StockType.CreateBox, PathStock.StockType.FromStock(stock))
        self.assertEqual(100, stock.Length)
        self.assertEqual(200, stock.Width)
        self.assertEqual(300, stock.Height)
        self.assertPlacement(orig.Placement, stock.Placement)

        # don't store placement in template
        template = PathStock.TemplateAttributes(orig, True, False)
        stock = PathStock.CreateFromTemplate(self.job, template)

        self.assertEqual(PathStock.StockType.CreateBox, PathStock.StockType.FromStock(stock))
        self.assertEqual(orig.Length, stock.Length)
        self.assertEqual(orig.Width, stock.Width)
        self.assertEqual(orig.Height, stock.Height)
        self.assertPlacement(FreeCAD.Placement(), stock.Placement)

    def test11(self):
        """Verify FromTemplate cylinder creation."""
        radius = 7
        height = 12
        placement = FreeCAD.Placement(FreeCAD.Vector(99, 88, 77), FreeCAD.Vector(1, 1, 1), 123)
        orig = PathStock.CreateCylinder(self.job, radius, height, placement)

        # full template
        template = PathStock.TemplateAttributes(orig)
        stock = PathStock.CreateFromTemplate(self.job, template)

        self.assertEqual(PathStock.StockType.CreateCylinder, PathStock.StockType.FromStock(stock))
        self.assertEqual(orig.Radius, stock.Radius)
        self.assertEqual(orig.Height, stock.Height)
        self.assertPlacement(orig.Placement, stock.Placement)

        # no extent in template
        template = PathStock.TemplateAttributes(orig, False, True)
        stock = PathStock.CreateFromTemplate(self.job, template)

        self.assertEqual(PathStock.StockType.CreateCylinder, PathStock.StockType.FromStock(stock))
        self.assertRoughly(R, stock.Radius.Value)
        self.assertEqual(300, stock.Height)
        self.assertPlacement(orig.Placement, stock.Placement)

        # no placement template - and no base
        template = PathStock.TemplateAttributes(orig, True, False)
        stock = PathStock.CreateFromTemplate(None, template)

        self.assertEqual(PathStock.StockType.CreateCylinder, PathStock.StockType.FromStock(stock))
        self.assertEqual(orig.Radius, stock.Radius)
        self.assertEqual(orig.Height, stock.Height)
        self.assertPlacement(FreeCAD.Placement(), stock.Placement)

        # no placement template - but base
        template = PathStock.TemplateAttributes(orig, True, False)
        stock = PathStock.CreateFromTemplate(self.job, template)

        self.assertEqual(PathStock.StockType.CreateCylinder, PathStock.StockType.FromStock(stock))
        self.assertEqual(orig.Radius, stock.Radius)
        self.assertEqual(orig.Height, stock.Height)
        self.assertPlacement(
            FreeCAD.Placement(FreeCAD.Vector(50, 100, 0), FreeCAD.Rotation()),
            stock.Placement,
        )

    def test12(self):
        """Verify FromTemplate from Base creation."""
        neg = FreeCAD.Vector(1, 2, 3)
        pos = FreeCAD.Vector(9, 8, 7)
        orig = PathStock.CreateFromBase(self.job, neg, pos)

        # Make sure we have a different base object for the creation
        self.base.Length = 11
        self.base.Width = 12
        self.base.Height = 13

        # full template
        template = PathStock.TemplateAttributes(orig)
        stock = PathStock.CreateFromTemplate(self.job, template)
        self.assertEqual(PathStock.StockType.FromBase, PathStock.StockType.FromStock(stock))
        self.assertEqual(orig.ExtXneg, stock.ExtXneg)
        self.assertEqual(orig.ExtXpos, stock.ExtXpos)
        self.assertEqual(orig.ExtYneg, stock.ExtYneg)
        self.assertEqual(orig.ExtYpos, stock.ExtYpos)
        self.assertEqual(orig.ExtZneg, stock.ExtZneg)
        self.assertEqual(orig.ExtZpos, stock.ExtZpos)

        bb = stock.Shape.BoundBox
        self.assertEqual(neg.x + pos.x + 11, bb.XLength)
        self.assertEqual(neg.y + pos.y + 12, bb.YLength)
        self.assertEqual(neg.z + pos.z + 13, bb.ZLength)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD material editor"
__author__ = "Yorik van Havre, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import os
from pathlib import PurePath
import sys
from PySide import QtCore, QtGui, QtSvgWidgets

import FreeCAD
import FreeCADGui
# import Material_rc
from materialtools.cardutils import get_material_preferred_directory, get_material_preferred_save_directory


# is this still needed after the move to card utils???
unicode = str


# ************************************************************************************************
# ************************************************************************************************
class MaterialEditor:

    def __init__(self, obj=None, prop=None, material=None, card_path="", category="Solid"):

        """Initializes, optionally with an object name and a material property
        name to edit, or directly with a material dictionary."""

        self.obj = obj
        self.prop = prop
        self.category = category
        self.material = material
        self.customprops = []
        self.internalprops = []
        self.groups = []
        self.directory = get_material_preferred_directory()
        self.save_directory = get_material_preferred_save_directory()
        if self.directory is None:
            self.directory = FreeCAD.getResourceDir() + "Mod/Material"
        self.materials = {}
        self.cards = {}
        self.icons = {}
        self.initialIndex = -1
        self.edited = False
        self.card_path = card_path
        filePath = os.path.dirname(__file__) + os.sep
        self.iconPath = (filePath + "Resources" + os.sep + "icons" + os.sep)

        # load the UI file from the same directory as this script
        self.widget = FreeCADGui.PySideUic.loadUi(":/ui/materials-editor.ui")
        # remove unused Help button
        self.widget.setWindowFlags(self.widget.windowFlags()
                                   & ~QtCore.Qt.WindowContextHelpButtonHint)

        # restore size and position
        param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material")
        width = param.GetInt("MaterialEditorWidth", 441)
        height = param.GetInt("MaterialEditorHeight", 626)
        self.widget.resize(width, height)

        # additional UI fixes and tweaks
        widget = self.widget
        buttonURL = widget.ButtonURL
        buttonDeleteProperty = widget.ButtonDeleteProperty
        buttonAddProperty = widget.ButtonAddProperty
        standardButtons = widget.standardButtons
        buttonOpen = widget.ButtonOpen
        buttonSave = widget.ButtonSave
        comboMaterial = widget.ComboMaterial
        treeView = widget.treeView

        # create preview svg slots
        self.widget.PreviewRender = QtSvgWidgets.QSvgWidget(self.iconPath + "preview-rendered.svg")
        self.widget.PreviewRender.setMaximumWidth(64)
        self.widget.PreviewRender.setMinimumHeight(64)
        self.widget.topLayout.addWidget(self.widget.PreviewRender)
        self.widget.PreviewVector = QtSvgWidgets.QSvgWidget(self.iconPath + "preview-vector.svg")
        self.widget.PreviewVector.setMaximumWidth(64)
        self.widget.PreviewVector.setMinimumHeight(64)
        self.widget.topLayout.addWidget(self.widget.PreviewVector)
        self.updatePreviews(mat=material)

        buttonURL.setIcon(QtGui.QIcon(":/icons/internet-web-browser.svg"))
        buttonDeleteProperty.setEnabled(False)
        standardButtons.button(QtGui.QDialogButtonBox.Ok).setAutoDefault(False)
        standardButtons.button(QtGui.QDialogButtonBox.Cancel).setAutoDefault(False)
        self.updateCardsInCombo()
        # TODO allow one to enter a custom property by pressing Enter in the lineedit
        # currently closes the dialog

        standardButtons.rejected.connect(self.reject)
        standardButtons.button(QtGui.QDialogButtonBox.Ok).clicked.connect(self.verify)
        buttonOpen.clicked.connect(self.openfile)
        buttonSave.clicked.connect(self.savefile)
        buttonURL.clicked.connect(self.openProductURL)
        comboMaterial.currentIndexChanged[int].connect(self.chooseMaterial)
        buttonAddProperty.clicked.connect(self.addCustomProperty)
        buttonDeleteProperty.clicked.connect(self.deleteCustomProperty)
        treeView.clicked.connect(self.onClickTree)

        model = QtGui.QStandardItemModel()
        treeView.setModel(model)
        treeView.setUniformRowHeights(True)
        treeView.setItemDelegate(MaterialsDelegate())
        model.itemChanged.connect(self.modelChange)

        # init model
        self.implementModel()

        # update the editor with the contents of the property, if we have one
        matProperty = None
        if self.prop and self.obj:
            matProperty = FreeCAD.ActiveDocument.getObject(self.obj).getPropertyByName(self.prop)
        elif self.material:
            matProperty = self.material

        if matProperty:
            self.updateMatParamsInTree(matProperty)
            self.widget.ComboMaterial.setCurrentIndex(0)
            # set after tree params to the none material

        if self.card_path:
            # we need the index of this path
            self.initialIndex = self.widget.ComboMaterial.findData(self.card_path)
            self.chooseMaterial(self.initialIndex)

        # TODO: What if material and card_name was given?
        # In such case ATM material is chosen, give some feedback for all those corner cases.

    def implementModel(self):

        """implements the model with the material attribute structure."""

        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material")
        widget = self.widget
        treeView = widget.treeView
        model = treeView.model()
        model.setHorizontalHeaderLabels(["Property", "Value", "Type"])

        treeView.setColumnWidth(0, 250)
        treeView.setColumnWidth(1, 250)
        treeView.setColumnHidden(2, True)

        from materialtools.cardutils import get_material_template
        template_data = get_material_template(True)

        for group in template_data:
            gg = list(group)[0]  # group dict has only one key
            top = QtGui.QStandardItem(gg)
            model.appendRow([top])
            self.groups.append(gg)

            for properName in group[gg]:
                pp = properName  # property name
                item = QtGui.QStandardItem(pp)
                item.setToolTip(group[gg][properName]["Description"])
                self.internalprops.append(pp)

                it = QtGui.QStandardItem()
                it.setToolTip(group[gg][properName]["Description"])

                tt = group[gg][properName]["Type"]
                itType = QtGui.QStandardItem(tt)

                top.appendRow([item, it, itType])
                treeView.setExpanded(top.index(), p.GetBool("TreeExpand"+gg, True))
            # top.sortChildren(0)

        # treeView.expandAll()
        self.edited = False

    def updateMatParamsInTree(self, data):

        """updates the contents of the editor with the given dictionary
           the material property keys where added to the editor already
           unknown material property keys will be added to the user defined group"""

        # print(data)
        model = self.widget.treeView.model()
        root = model.invisibleRootItem()
        for gg in range(root.rowCount() - 1):
            group = root.child(gg, 0)
            for pp in range(group.rowCount()):
                item = group.child(pp, 0)
                it = group.child(pp, 1)
                kk = self.collapseKey(item.text())

                try:
                    value = data[kk]
                    it.setText(value)
                    del data[kk]
                except KeyError:
                    # treat here changes in Material Card Template
                    # Norm -> StandardCode
                    if (kk == "Standard Code") and ("Norm" in data) and data["Norm"]:
                        it.setText(data["Norm"])
                        del data["Norm"]
                    it.setText("")

        userGroup = root.child(gg + 1, 0)
        userGroup.setRowCount(0)
        self.customprops = []

        for k, i in data.items():
            k = self.expandKey(k)
            item = QtGui.QStandardItem(k)
            it = QtGui.QStandardItem(i)
            userGroup.appendRow([item, it])
            self.customprops.append(k)
        self.edited = False

    def chooseMaterial(self, index):
        if index < 0:
            return

        if self.verifyMaterial():
            """
                Save any unchanged data
            """
            self.edited = False
        else:
            return

        self.card_path = self.widget.ComboMaterial.itemData(index)
        FreeCAD.Console.PrintMessage(
            "choose_material in material editor:\n"
            "    {}\n".format(self.card_path)
        )
        if os.path.isfile(self.card_path):
            from importFCMat import read
            d = read(self.card_path)
            self.updateMatParamsInTree(d)
            # be careful with reading from materials dict
            # the card could be updated the dict not
            self.widget.ComboMaterial.setCurrentIndex(index)  # set after tree params
        else:
            FreeCAD.Console.PrintError("Material card not found: {}\n".format(self.card_path))

    def verifyMaterial(self):
        if self.edited:
            reply = QtGui.QMessageBox.question(self.widget, #FreeCADGui.getMainWindow(),
                                                translate("Material","The document has been modified."),
                                                translate("Material","Save changes?"),
                                                QtGui.QMessageBox.Save | QtGui.QMessageBox.Discard | QtGui.QMessageBox.Cancel,
                                                QtGui.QMessageBox.Save)

            if reply == QtGui.QMessageBox.Cancel:
                return False
            if reply == QtGui.QMessageBox.Save:
                self.savefile()

        return True

    def updateCardsInCombo(self):

        """updates the contents of the materials combo with existing material cards"""

        mat_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material/Cards")
        sort_by_resources = mat_prefs.GetBool("SortByResources", False)

        # get all available materials (fill self.materials, self.cards and self.icons)
        from materialtools.cardutils import import_materials as getmats
        self.materials, self.cards, self.icons = getmats(category=self.category)

        card_name_list = []  # [ [card_name, card_path, icon_path], ... ]

        if sort_by_resources is True:
            for a_path in sorted(self.materials):
                card_name_list.append([self.cards[a_path], a_path, self.icons[a_path]])
        else:
            card_names_tmp = {}
            for path, name in self.cards.items():
                card_names_tmp[name] = path
            for a_name in sorted(card_names_tmp):
                a_path = card_names_tmp[a_name]
                card_name_list.append([a_name, a_path, self.icons[a_path]])

        card_name_list.insert(0, [None, "", ""])
        self.widget.ComboMaterial.clear()
        for mat in card_name_list:
            icon_data = mat[2]

            if icon_data and isinstance(icon_data, bytes):
                byte_array = QtCore.QByteArray(icon_data)
                pixmap = QtGui.QPixmap()
                if pixmap.loadFromData(byte_array):
                    icon = QtGui.QIcon(pixmap)
                else:
                    icon = QtGui.QIcon()
            elif isinstance(icon_data, str) and icon_data:
                # if this is string type, then try to load directly
                icon = QtGui.QIcon(icon_data)
            else:
                # fallback to not crash, empty icon
                icon = QtGui.QIcon()

            self.widget.ComboMaterial.addItem(icon, mat[0], mat[1])

    def openProductURL(self):

        """opens the contents of the ProductURL field in an external browser."""

        model = self.widget.treeView.model()
        item = model.findItems(translate("Material", "Product URL"),
                               QtCore.Qt.MatchRecursive, 0)[0]
        group = item.parent()
        it = group.child(item.row(), 1)
        url = it.text()
        if url:
            QtGui.QDesktopServices.openUrl(QtCore.QUrl(url, QtCore.QUrl.TolerantMode))

    def modelChange(self, item):
        """
            Called when an item in the tree is modified. This will set edited to True, but this
            will be reset in the event of mass updates, such as loading a card
        """
        self.edited = True

    def verify(self, button):
        """
            Verify that the user wants to save any changed data before exiting
        """

        if self.edited:
            reply = QtGui.QMessageBox.question(self.widget, #FreeCADGui.getMainWindow(),
                                                translate("Material","The document has been modified."),
                                                translate("Material","Save changes?"),
                                                QtGui.QMessageBox.Save | QtGui.QMessageBox.Discard | QtGui.QMessageBox.Cancel,
                                                QtGui.QMessageBox.Save)

            if reply == QtGui.QMessageBox.Cancel:
                return
            if reply == QtGui.QMessageBox.Save:
                self.savefile()

        self.accept()

    def accept(self):
        ""

        self.storeSize()
        QtGui.QDialog.accept(self.widget)

    def reject(self):
        ""

        self.storeSize()
        QtGui.QDialog.reject(self.widget)

    def storeSize(self):
        "stores the widget size"
        # store widths
        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material")
        p.SetInt("MaterialEditorWidth", self.widget.width())
        p.SetInt("MaterialEditorHeight", self.widget.height())
        root = self.widget.treeView.model().invisibleRootItem()
        for gg in range(root.rowCount()):
            group = root.child(gg)
            p.SetBool("TreeExpand"+group.text(), self.widget.treeView.isExpanded(group.index()))

    def expandKey(self, key):
        "adds spaces before caps in a KeyName"
        nk = ""
        for ln in key:
            if ln.isupper():
                if nk:
                    # this allows for series of caps, such as ProductURL
                    if not nk[-1].isupper():
                        nk += " "
            nk += ln
        return nk

    def collapseKey(self, key):
        "removes the spaces in a Key Name"
        nk = ""
        for ln in key:
            if ln != " ":
                nk += ln
        return nk

    def addCustomProperty(self, key=None, value=None):
        "Adds a custom property to the editor, optionally with a value."

        if not key:
            key = self.widget.EditProperty.text()

        if key:
            model = self.widget.treeView.model()
            item = model.findItems(key, QtCore.Qt.MatchRecursive, 0)
            if not item:

                top = model.findItems(translate("Material", "UserDefined"),
                                      QtCore.Qt.MatchExactly, 0)[0]
                item = QtGui.QStandardItem(key)
                it = QtGui.QStandardItem(value)
                top.appendRow([item, it])
                self.customprops.append(key)
                self.edited = True

    def deleteCustomProperty(self, key=None):

        """Deletes a custom property from the editor,
        or deletes the value of an internal property."""

        widget = self.widget
        treeView = widget.treeView
        model = treeView.model()
        buttonDeleteProperty = widget.ButtonDeleteProperty

        if not key:

            index = treeView.selectedIndexes()[0]
            item = model.itemFromIndex(index)
            key = item.text()

        if key:
            item = model.findItems(key, QtCore.Qt.MatchRecursive, 0)
            if item:

                index = model.indexFromItem(item[0])
                topIndex = index.parent()
                top = model.itemFromIndex(topIndex)
                row = item[0].row()

                if key in self.customprops:
                    top.takeRow(row)
                    self.customprops.remove(key)
                    buttonDeleteProperty.setProperty("text", "Delete property")

                elif key in self.internalprops:
                    it = top.child(row, 1)
                    it.setText("")
                    buttonDeleteProperty.setProperty("text", "Delete value")

                self.edited = True

        buttonDeleteProperty.setEnabled(False)

    def onClickTree(self, index):

        """Checks if the current item is a custom or an internal property,
        and enable the delete property or delete value button."""

        widget = self.widget
        buttonDeleteProperty = widget.ButtonDeleteProperty
        treeView = widget.treeView
        model = treeView.model()
        ind = treeView.selectedIndexes()[0]
        item = model.itemFromIndex(ind)
        text = item.text()

        if text in self.customprops:
            buttonDeleteProperty.setEnabled(True)
            buttonDeleteProperty.setProperty("text", "Delete property")

        elif text in self.internalprops:
            indParent = ind.parent()
            group = model.itemFromIndex(indParent)
            row = item.row()
            it = group.child(row, 1)
            buttonDeleteProperty.setProperty("text", "Delete value")
            if it.text():
                buttonDeleteProperty.setEnabled(True)
            else:
                buttonDeleteProperty.setEnabled(False)

        else:
            buttonDeleteProperty.setEnabled(False)
            buttonDeleteProperty.setProperty("text", "Delete property")

        self.updatePreviews()

    def getDict(self):
        "returns a dictionary from the contents of the editor."

        model = self.widget.treeView.model()
        if model is None:
            return {}
        root = model.invisibleRootItem()

        d = {}
        for gg in range(root.rowCount()):
            group = root.child(gg)
            for row in range(group.rowCount()):
                kk = group.child(row, 0).text()
                ii = group.child(row, 1).text()

                # TODO the following should be translated back to english
                # since text(0) could be translated
                matkey = self.collapseKey(str(kk))
                matvalue = unicode(ii)
                if matvalue or (matkey == "Name"):
                    # use only keys which are not empty and the name even if empty
                    d[matkey] = matvalue
        # self.outputDict(d)
        return d

    def outputDict(self, d):
        print("MaterialEditor dictionary")
        for param in d:
            print("  {} : {}".format(param, d[param]))

    """
    def setTexture(self, pattern):
        "displays a texture preview if needed"
        self.widget.PreviewVector.hide()
        if pattern:
            try:
                import DrawingPatterns
            except:
                print("DrawingPatterns not found")
            else:
                pattern = DrawingPatterns.buildFileSwatch(pattern, size=96, png=True)
                if pattern:
                    self.widget.PreviewVector.setPixmap(QtGui.QPixmap(pattern))
                    self.widget.PreviewVector.show()
    """

    def updatePreviews(self, mat=None):
        "updates the preview images from the content of the editor"
        if not mat:
            mat = self.getDict()
        diffcol = None
        highlightcol = None
        sectioncol = None
        if "DiffuseColor" in mat:
            diffcol = mat["DiffuseColor"]
        elif "ViewColor" in mat:
            diffcol = mat["ViwColor"]
        elif "Color" in mat:
            diffcol = mat["Color"]
        if "SpecularColor" in mat:
            highlightcol = mat["SpecularColor"]
        if "SectionColor" in mat:
            sectioncol = mat["SectionColor"]
        if diffcol or highlightcol:
            fd = QtCore.QFile(self.iconPath + "preview-rendered.svg")
            if fd.open(QtCore.QIODevice.ReadOnly | QtCore.QIODevice.Text):
                svg = QtCore.QTextStream(fd).readAll()
                fd.close()
                if diffcol:
                    svg = svg.replace("#d3d7cf", self.getColorHash(diffcol, val=255))
                    svg = svg.replace("#555753", self.getColorHash(diffcol, val=125))
                if highlightcol:
                    svg = svg.replace("#fffffe", self.getColorHash(highlightcol, val=255))
                self.widget.PreviewRender.load(QtCore.QByteArray(bytes(svg, encoding="utf8")))
        if diffcol or sectioncol:
            fd = QtCore.QFile(self.iconPath + "preview-vector.svg")
            if fd.open(QtCore.QIODevice.ReadOnly | QtCore.QIODevice.Text):
                svg = QtCore.QTextStream(fd).readAll()
                fd.close()
                if diffcol:
                    svg = svg.replace("#d3d7cf", self.getColorHash(diffcol, val=255))
                    svg = svg.replace("#555753", self.getColorHash(diffcol, val=125))
                if sectioncol:
                    svg = svg.replace("#ffffff", self.getColorHash(sectioncol, val=255))
                self.widget.PreviewVector.load(QtCore.QByteArray(bytes(svg, encoding="utf8")))

    def getColorHash(self, col, val=255):
        "returns a '#000000' string from a '(0.1,0.2,0.3)' string"
        col = [float(x.strip()) for x in col.strip("()").split(",")]
        color = QtGui.QColor(int(col[0]*val), int(col[1]*val), int(col[2]*val))
        return color.name()

    def openfile(self):
        if self.verifyMaterial():
            """
                Save any unchanged data
            """
            self.edited = False
        else:
            return

        "Opens a FCMat file"
        if self.category == "Solid":
            directory = self.directory + os.sep + "StandardMaterial"
        else:
            directory = self.directory + os.sep + "FluidMaterial"
        if self.card_path is None or len(self.card_path) == 0:
            self.card_path = directory
        filetuple = QtGui.QFileDialog.getOpenFileName(
            QtGui.QApplication.activeWindow(),
            "Open FreeCAD material file",
            self.card_path,
            "*.FCMat"
        )
        if os.path.isfile(filetuple[0]):
            card_path_new = filetuple[0]
        else:
            FreeCAD.Console.PrintError("Error: Invalid path to the material file\n")
            return
        # we cannot simply execute findData(self.card_path) because e.g. on Windows
        # the return path has "/" as folder separator, but the paths in the ComboMaterial
        # have also some "\" in them. For example a path can look like this:
        # D:/FreeCAD-build/data/Mod\Material\FluidMaterial\Air.FCMat
        # To keep it simple, we take a path from the ComboMaterial and change only the
        # material card filename
        #
        # Using the initialIndex variable won't work before a card os selected for the
        # first time, so use index 1. Index 0 is a blank entry
        if self.widget.ComboMaterial.count() > 1:
            path = self.widget.ComboMaterial.itemData(1)
            # at first check if we have a uniform usage
            # (if a character is not present, rsplit delivers the initial string)
            testBackslash = path.rsplit('\\', 1)[0]
            testSlash = path.rsplit('/', 1)[0]
            if testBackslash == path:
                path = testBackslash.rsplit('/', 1)[0] + '/'
            elif testSlash == path:
                path = testSlash.rsplit('\\', 1)[0] + '\\'
            # since we don't know if we have to deal with slash or backslash, take the
            # longest result as path
            else:
                pathBackslash = path.rsplit('\\', 1)[0]
                pathSlash = path.rsplit('/', 1)[0]
                if len(pathBackslash) > len(pathSlash):
                    path = pathBackslash + '\\'
                else:
                    path = pathSlash + '/'
            # we know that card_path_new has uniformly either / or \ but not yet what
            testBackslash = card_path_new.rsplit('\\', 1)[0]
            if testBackslash == card_path_new:
                self.card_path = path + card_path_new.rsplit('/', 1)[1]
            else:
                self.card_path = path + card_path_new.rsplit('\\', 1)[1]
        index = self.widget.ComboMaterial.findData(self.card_path)

        # check if card_path is in known path, means it is in combo box already
        # if not print message, and give some feedbach that the card parameter are loaded
        if index == -1:
            FreeCAD.Console.PrintMessage(
                "Card path: {} not found in known cards.\n"
                "The material parameter only are loaded.\n"
                .format(self.card_path)
            )
            # a material card was chosen that is unknown, thus use its full path
            self.card_path = card_path_new
            from importFCMat import read
            materialDict = read(self.card_path)
            if materialDict:
                self.updateMatParamsInTree(materialDict)
                self.widget.ComboMaterial.setCurrentIndex(0)
                # set combo box to the none material after tree params
        else:
            self.chooseMaterial(index)
        self.directory = os.path.dirname(self.card_path)

    def savefile(self):
        "Saves a FCMat file."

        model = self.widget.treeView.model()
        item = model.findItems(translate("Material", "Name"),
                               QtCore.Qt.MatchRecursive, 0)[0]
        group = item.parent()
        it = group.child(item.row(), 1)
        name = it.text()
        if not name:
            name = "Material"
        filetuple = QtGui.QFileDialog.getSaveFileName(
            QtGui.QApplication.activeWindow(),
            "Save FreeCAD material file",
            self.save_directory + "/" + name + ".FCMat",
            "*.FCMat"
        )
        # a tuple of two empty strings returns True, so use the filename directly
        filename = filetuple[0]
        if filename:
            # Update the directories to the current save value
            self.save_directory = os.path.dirname(filename)
            self.directory = self.save_directory
            self.card_path = filename

            d = self.getDict()
            # self.outputDict(d)
            if d:
                # Set the card name to match the filename
                path = PurePath(filename)
                d["CardName"] = path.stem

                from importFCMat import write
                write(filename, d)
                import Materials
                # Load the material
                manager = Materials.MaterialManager()
                manager.getMaterialByPath(filename)
                self.edited = False
                self.updateCardsInCombo()

                # Ensure our card is selected
                self.widget.ComboMaterial.setCurrentText(path.stem)

    def show(self):
        return self.widget.show()

    def exec_(self):
        return self.widget.exec_()


# ************************************************************************************************
# ************************************************************************************************
class MaterialsDelegate(QtGui.QStyledItemDelegate):

    """provides display and editing facilities for data items from a model."""

    def __init__(self):
        ""

        self.matproperty = ""
        self.Value = ""
        super(MaterialsDelegate, self).__init__()

    def createEditor(self, parent, option, index):

        """returns the widget used to change data from the model."""

        model = index.model()
        column = index.column()

        item = model.itemFromIndex(index)
        group = item.parent()
        if not group:
            return

        if column == 1:
            row = index.row()

            PP = group.child(row, 0)
            self.matproperty = PP.text().replace(" ", "")  # remove spaces

            TT = group.child(row, 2)
            if TT:
                Type = TT.text()
            else:
                Type = "String"

            VV = group.child(row, 1)
            self.Value = VV.text()

            editor = matProperWidget(parent, self.matproperty, Type, self.Value)

        elif column == 0:
            if group.text() == "User defined":
                editor = matProperWidget(parent)
            else:
                return

        else:
            return

        return editor

    def setEditorData(self, editor, index):

        """provides the widget with data to manipulate."""

        Type = editor.property("Type")
        model = index.model()
        item = model.itemFromIndex(index)

        if Type == "Color":
            color = editor.property("color")
            color = tuple([v/255.0 for v in color.getRgb()])
            item.setText(str(color))

        elif Type == "File":
            lineEdit = editor.children()[1]
            item.setText(lineEdit.text())

        elif Type == "Float":
            # avoid rounding artifacts
            inputValue = float('%.6g' % editor.value())
            item.setText(str(inputValue))

        elif Type == "Quantity":
            if not hasattr(FreeCAD.Units, self.matproperty):
                FreeCAD.Console.PrintError(
                    "Error: property '{}' is a quantity but has no unit defined\n"
                    .format(self.matproperty)
                )
                return
            # we must use the unit of the input value because the
            # "Gui::QuantitySpinBox" uses e.g. for the density always the mm-based unit
            # kg/mm^3, also when the input value is in kg/^3.
            # E.g. when the input is e.g. "7875 kg/m^3" and we would pass "7875" as rawValue
            # and "kg/m^3" as unit, we would get "7875 kg/mm^3" as result. If we try to be
            # clever and input "7875e-6" as rawValue and "kg/m^3" as unit we get
            # "7875e-6 kg/m^3" as result. If we input "7875e-6" as rawValue and "kg/mm^3"
            # as unit we get also "7875e-6 kg/m^3" as result.
            if not self.Value:
                # for empty (not yet set properties) we use the matproperty unit
                unit = getattr(FreeCAD.Units, self.matproperty)
                quantity = FreeCAD.Units.Quantity(1, unit)
                item.setText(str(editor.value()) + " " + quantity.getUserPreferred()[2])
            else:
                # since we checked we have a quantity, we can use split() to get the unit
                item.setText(str(editor.value()) + " " + self.Value.split()[1])

        else:
            super(MaterialsDelegate, self).setEditorData(editor, index)


# ************************************************************************************************
# ************************************************************************************************
def matProperWidget(parent=None, matproperty=None, Type="String", Value=None,
                    minimum=None, maximum=None, stepsize=None, precision=12):

    """customs widgets for the material stuff."""

    # FIXME
    # Workaround for problem from here:
    # https://forum.freecad.org/viewtopic.php?f=18&t=56912&start=20#p516811
    # set precision to 12
    # Better would be a similar system like used in FEM material task panel
    # if the value in the InputField has not changed the data is not changed
    # why does the value changes if the user clicks in
    # the value and unit should exact stay as it is displayed before the user
    # clicks in the field

    # the user defined properties are of Type String and thus uses a "Gui::PrefLineEdit"

    ui = FreeCADGui.UiLoader()

    if Type == "String":
        widget = ui.createWidget("Gui::PrefLineEdit")

    elif Type == "URL":
        widget = ui.createWidget("Gui::PrefLineEdit")

    elif Type == "File":
        widget = ui.createWidget("Gui::FileChooser")
        if Value:
            lineEdit = widget.children()[1]
            lineEdit.setText(Value)

    elif Type == "Quantity":
        # We don't use a Gui::QuantitySpinBox widget because depending on the value,
        # the unit might change. For some inputs there is no way to do this right,
        # see the comment above in "def setEditorData". Moreover it is error-prone to provide
        # a unit as in the Gui::QuantitySpinBox widget because users likely delete the unit or
        # change it accidentally. It is therefore better not to display the unit while editing.
        widget = ui.createWidget("Gui::DoubleSpinBox")

        if minimum is None:
            widget.setMinimum(-1*sys.float_info.max)
        else:
            widget.setMinimum(minimum)
        if maximum is None:
            widget.setMaximum(sys.float_info.max)
        else:
            widget.setMaximum(maximum)

        # we must increase the digits before we can set the value
        # 6 is sufficient as some metarial cards use e.g. "0.000011"
        widget.setDecimals(6)

        # for properties with an underscored number (vectorial values),
        # we must strip the part after the first underscore to obtain the bound unit
        # since in cardutils.py in def get_material_template
        # the underscores were removed, we must check for numbers
        import re
        if re.search(r'\d', matproperty):
            matpropertyNum = matproperty.rstrip('0123456789')
            matproperty = matpropertyNum

        if Value:
            widget.setValue(float(Value.split()[0]))

    elif Type == "Integer":
        widget = ui.createWidget("Gui::UIntSpinBox")
        if minimum is None:
            widget.setMinimum(0)
        else:
            widget.setMinimum(minimum)
        if maximum is None:
            widget.setMaximum(sys.maxsize)
        else:
            widget.setMaximum(maximum)

    elif Type == "Float":
        widget = ui.createWidget("Gui::DoubleSpinBox")
        # the magnetic permeability is the parameter for which many decimals matter
        # the most however, even for this, 6 digits are sufficient
        widget.setDecimals(6)
        # for almost all Float parameters of materials a step of 1 would be too large
        widget.setSingleStep(0.1)
        if minimum is None:
            widget.setMinimum(sys.float_info.min)
        else:
            widget.setMinimum(minimum)
        if maximum is None:
            widget.setMaximum(sys.float_info.max)
        else:
            widget.setMaximum(maximum)
        widget.setValue(float(Value))

    elif Type == "Enumerator":
        widget = ui.createWidget("Gui::PrefComboBox")

    elif Type == "Boolean":
        widget = ui.createWidget("Gui::PrefComboBox")
        widget.insertItems(0, ["", "False", "True"])

    elif Type == "Vector":
        widget = ui.createWidget("Gui::PrefLineEdit")

    elif Type == "Color":
        widget = ui.createWidget("Gui::PrefColorButton")
        if Value:
            value = string2tuple(Value)
            color = QtGui.QColor()
            color.setRgb(value[0], value[1], value[2], value[3])
            widget.setProperty("color", color)

    else:
        widget = QtGui.QLineEdit()

    widget.setProperty("Type", Type)

    widget.setParent(parent)

    return widget


def string2tuple(string):
    "provisionally"
    value = string[1:-1]
    value = value.split(",")
    value = [int(float(v)*255) for v in value]
    value = tuple(value)
    return value


def translate(context, text):
    "translates text"
    return text  # TODO use Qt translation mechanism here


def openEditor(obj=None, prop=None):
    """openEditor([obj,prop]): opens the editor, optionally with
    an object name and material property name to edit"""
    # param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Material/Cards")
    # legacy = param.GetBool("LegacyEditor", True)
    # if legacy:
    #     editor = MaterialEditor(obj, prop)
    #     editor.exec_()
    # else:
    FreeCADGui.runCommand('Material_Edit',0)


def editMaterial(material=None, card_path=None, category="Solid"):
    """editMaterial(material): opens the editor to edit the contents
    of the given material dictionary. Returns the modified material dictionary."""
    # if the material editor is opened with this def and the card_path is None
    # the combo box with the card name is empty
    # this makes sense, because the editor was not opened with a card_path,
    # but with material dictionary instead
    # TODO: add some text in combo box, may be "custom material data" or "user material data"
    # TODO: if card_path is None, all known cards could be checked,
    # if one fits exact ALL provided data, this card name could be displayed
    editor = MaterialEditor(material=material, card_path=card_path, category=category)
    result = editor.exec_()
    if result:
        return editor.getDict()
    else:
        # on cancel button an empty dict is returned
        return {}


# ************************************************************************************************
# ************************************************************************************************
"""
# some examples how to open the material editor in Python:
import MaterialEditor
MaterialEditor.openEditor()

doc = FreeCAD.open(
    FreeCAD.ConfigGet("AppHomePath") + "data/examples/FEMExample.FCStd"
)
import MaterialEditor
MaterialEditor.openEditor("SolidMaterial", "Material")

import MaterialEditor
MaterialEditor.editMaterial({
    "Density": "1234.0 kg/m^3",
    "Name": "My-Material-Data",
    "PoissonRatio": "0.66",
    "YoungsModulus": "123456 MPa"
})

import MaterialEditor
MaterialEditor.editMaterial("ABS")

"""



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Mystran add fem mesh"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from feminout import exportNastranMesh
from femmesh import meshtools


def add_mesh(f, model, mystran_writer):

    # needed basic data
    if not mystran_writer.femnodes_mesh:
        mystran_writer.femnodes_mesh = mystran_writer.femmesh.Nodes
    if not mystran_writer.femelement_table:
        mystran_writer.femelement_table = meshtools.get_femelement_table(mystran_writer.femmesh)
    mesh_eletype = exportNastranMesh.get_export_element_type(
        mystran_writer.femmesh, mystran_writer.femelement_table
    )

    # get the pynas code
    mesh_pynas_code = exportNastranMesh.get_pynastran_mesh(
        mystran_writer.femnodes_mesh, mystran_writer.femelement_table, mesh_eletype
    )
    # print(mesh_pynas_code)

    # write the pyNastran code
    f.write(mesh_pynas_code)

    # execute pyNastran code to add grid to the model
    # print(model.get_bdf_stats())
    exec(mesh_pynas_code)
    # print(model.get_bdf_stats())

    return model


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# Unit tests for the Path.Tool.Shape module and its utilities.

from pathlib import Path
from typing import Mapping, Tuple
import FreeCAD
from CAMTests.PathTestUtils import PathTestWithAssets
from Path.Tool.assets import DummyAssetSerializer
from Path.Tool.shape import (
    ToolBitShape,
    ToolBitShapeBallend,
    ToolBitShapeVBit,
    ToolBitShapeBullnose,
    ToolBitShapeSlittingSaw,
)


# Helper dummy class for testing abstract methods
class DummyShape(ToolBitShape):
    name = "dummy"

    def __init__(self, id, **kwargs):
        super().__init__(id=id, **kwargs)
        # Always define defaults in the subclass
        self._defaults = {
            "Param1": FreeCAD.Units.Quantity("10 mm"),
            "Param2": FreeCAD.Units.Quantity("5 deg"),
        }
        # Merge defaults into _params, allowing kwargs to override
        self._params = self._defaults | self._params

    @classmethod
    def schema(cls) -> Mapping[str, Tuple[str, str]]:
        return {
            "Param1": (
                FreeCAD.Qt.translate("Param1", "Parameter 1"),
                "App::PropertyLength",
            ),
            "Param2": (
                FreeCAD.Qt.translate("Param2", "Parameter 2"),
                "App::PropertyAngle",
            ),
        }

    @property
    def label(self):
        return "Dummy Shape"


def unit(param):
    return param.getUserPreferred()[2]


class TestPathToolShapeClasses(PathTestWithAssets):
    """Tests for the concrete ToolBitShape subclasses."""

    def _test_shape_common(self, alias):
        uri = ToolBitShape.resolve_name(alias)
        shape = self.assets.get(uri)
        return shape.get_parameters()

    def test_base_init_with_defaults(self):
        """Test base class initialization uses default parameters."""
        # Provide a dummy filepath and id for instantiation
        shape = DummyShape(id="dummy_shape_1", filepath=Path("/fake/dummy.fcstd"))
        self.assertEqual(shape.get_parameter("Param1").Value, 10.0)
        self.assertEqual(shape.get_parameter("Param1").Unit, FreeCAD.Units.Unit("mm"))
        self.assertEqual(shape.get_parameter("Param2").Value, 5.0)
        self.assertEqual(shape.get_parameter("Param2").Unit, FreeCAD.Units.Unit("deg"))

    def test_base_init_with_kwargs(self):
        """Test base class initialization overrides defaults with kwargs."""
        # Provide a dummy filepath for instantiation
        shape = DummyShape(
            id="dummy_shape_2",
            filepath=Path("/fake/dummy.fcstd"),
            Param1=FreeCAD.Units.Quantity("20 mm"),
            Param3="Ignored",
        )
        self.assertEqual(shape.get_parameter("Param1").Value, 20.0)
        self.assertEqual(shape.get_parameter("Param1").Value, 20.0)
        self.assertEqual(
            str(shape.get_parameter("Param1").Unit),
            "Unit: mm (1,0,0,0,0,0,0,0) [Length]",
        )
        self.assertEqual(shape.get_parameter("Param2").Value, 5.0)
        self.assertEqual(
            str(shape.get_parameter("Param2").Unit),
            "Unit: deg (0,0,0,0,0,0,0,1) [Angle]",
        )  # Should remain default

    def test_base_get_set_parameter(self):
        """Test getting and setting individual parameters."""
        # Provide a dummy filepath for instantiation
        shape = DummyShape(id="dummy_shape_3", filepath=Path("/fake/dummy.fcstd"))
        self.assertEqual(shape.get_parameter("Param1").Value, 10.0)
        self.assertEqual(shape.get_parameter("Param1").Unit, FreeCAD.Units.Unit("mm"))
        shape.set_parameter("Param1", FreeCAD.Units.Quantity("15 mm"))
        self.assertEqual(shape.get_parameter("Param1").Value, 15.0)
        self.assertEqual(shape.get_parameter("Param1").Unit, FreeCAD.Units.Unit("mm"))
        with self.assertRaisesRegex(KeyError, "Shape 'dummy' has no parameter 'InvalidParam'"):
            shape.get_parameter("InvalidParam")

    def test_base_get_parameters(self):
        """Test getting the full parameter dictionary."""
        # Provide a dummy filepath for instantiation
        shape = DummyShape(
            id="dummy_shape_4",
            filepath=Path("/fake/dummy.fcstd"),
            Param1=FreeCAD.Units.Quantity("12 mm"),
        )
        # Create mock quantity instances using the configured mock class
        expected_param1 = FreeCAD.Units.Quantity("12.0 mm")
        expected_param2 = FreeCAD.Units.Quantity("5.0 deg")

        expected = {"Param1": expected_param1, "Param2": expected_param2}
        params = shape.get_parameters()
        self.assertEqual(params["Param1"].Value, expected["Param1"].Value)
        self.assertEqual(str(params["Param1"].Unit), str(expected["Param1"].Unit))
        self.assertEqual(params["Param2"].Value, expected["Param2"].Value)
        self.assertEqual(str(params["Param2"].Unit), str(expected["Param2"].Unit))

    def test_base_name_property(self):
        """Test the name property returns the primary alias."""
        # Provide a dummy filepath for instantiation
        shape = DummyShape(id="dummy_shape_5", filepath=Path("/fake/dummy.fcstd"))
        self.assertEqual(shape.name, "dummy")

    def test_base_get_parameter_label(self):
        """Test retrieving parameter labels."""
        # Provide a dummy filepath for instantiation
        shape = DummyShape(id="dummy_shape_6", filepath=Path("/fake/dummy.fcstd"))
        self.assertEqual(shape.get_parameter_label("Param1"), "Parameter 1")
        self.assertEqual(shape.get_parameter_label("Param2"), "Parameter 2")
        # Test fallback for unknown parameter
        self.assertEqual(shape.get_parameter_label("UnknownParam"), "UnknownParam")

    def test_base_get_expected_shape_parameters(self):
        """Test retrieving the list of expected parameter names."""
        expected = ["Param1", "Param2"]
        self.assertCountEqual(DummyShape.get_expected_shape_parameters(), expected)

    def test_base_str_repr(self):
        """Test string representation."""
        # Provide a dummy filepath for instantiation
        shape = DummyShape(id="dummy_shape_7", filepath=Path("/fake/dummy.fcstd"))
        # Dynamically construct the expected string using the actual parameter string representations
        params_str = ", ".join(f"{name}={str(val)}" for name, val in shape.get_parameters().items())
        expected_str = f"dummy({params_str})"
        self.assertEqual(str(shape), expected_str)
        self.assertEqual(repr(shape), expected_str)

    def test_base_resolve_name(self):
        """Test resolving shape aliases to canonical names."""
        self.assertEqual(ToolBitShape.resolve_name("ballend").asset_id, "ballend")
        self.assertEqual(ToolBitShape.resolve_name("v-bit").asset_id, "v-bit")
        self.assertEqual(ToolBitShape.resolve_name("vbit").asset_id, "vbit")
        self.assertEqual(ToolBitShape.resolve_name("bullnose").asset_id, "bullnose")
        self.assertEqual(ToolBitShape.resolve_name("bullnose.fcstd").asset_id, "bullnose")
        self.assertEqual(ToolBitShape.resolve_name("SlittingSaw").asset_id, "SlittingSaw")
        # Test unknown name - should return the input name
        self.assertEqual(ToolBitShape.resolve_name("nonexistent").asset_id, "nonexistent")
        self.assertEqual(ToolBitShape.resolve_name("UnknownShape").asset_id, "UnknownShape")

    def test_concrete_classes_instantiation(self):
        """Test that all concrete classes can be instantiated."""
        # No patching of FreeCAD document operations here.
        # The test relies on the actual FreeCAD environment.

        shape_uris = self.assets.list_assets(asset_type="toolbitshape")
        for uri in shape_uris:
            # Skip the DummyShape asset if it exists
            if uri.asset_id == "dummy":
                continue

            with self.subTest(uri=uri):
                instance = self.assets.get(uri)
                self.assertIsInstance(instance, ToolBitShape)
                # Check if default params were set by checking if the
                # parameters dictionary is not empty.
                self.assertTrue(instance.get_parameters())

    def test_get_shape_class(self):
        """Test the get_shape_class function."""
        uri = ToolBitShape.resolve_name("ballend")
        self.assets.get(uri)  # Ensure it's loadable

        self.assertEqual(ToolBitShape.get_subclass_by_name("ballend"), ToolBitShapeBallend)
        self.assertEqual(ToolBitShape.get_subclass_by_name("v-bit"), ToolBitShapeVBit)
        self.assertEqual(ToolBitShape.get_subclass_by_name("VBit"), ToolBitShapeVBit)
        self.assertEqual(ToolBitShape.get_subclass_by_name("torus"), ToolBitShapeBullnose)
        self.assertEqual(ToolBitShape.get_subclass_by_name("slitting-saw"), ToolBitShapeSlittingSaw)
        self.assertIsNone(ToolBitShape.get_subclass_by_name("nonexistent"))

    # The following tests for default parameters and labels
    # should also not use mocks for FreeCAD document operations or Units.
    # They should rely on the actual FreeCAD environment and the
    # load_file method of the base class.

    def test_toolbitshapeballend_defaults(self):
        """Test ToolBitShapeBallend default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        # The actual file content is not loaded in this test,
        # only the default parameters are checked.
        shape = self._test_shape_common("ballend")
        self.assertEqual(shape["Diameter"].Value, 5.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["Length"].Value, 50.0)
        self.assertEqual(unit(shape["Length"]), "mm")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("ballend")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("Diameter"), "Diameter")
        self.assertEqual(instance.get_parameter_label("Length"), "Overall tool length")

    def test_toolbitshapedrill_defaults(self):
        """Test ToolBitShapeDrill default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("drill")
        self.assertEqual(shape["Diameter"].Value, 3.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["TipAngle"].Value, 119.0)
        self.assertEqual(unit(shape["TipAngle"]), "°")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("drill")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("Diameter"), "Diameter")
        self.assertEqual(instance.get_parameter_label("TipAngle"), "Tip angle")

    def test_toolbitshapechamfer_defaults(self):
        """Test ToolBitShapeChamfer default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("chamfer")
        self.assertEqual(shape["Diameter"].Value, 12.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["CuttingEdgeAngle"].Value, 60.0)
        self.assertEqual(unit(shape["CuttingEdgeAngle"]), "°")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("chamfer")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("Diameter"), "Diameter")

    def test_toolbitshapedovetail_defaults(self):
        """Test ToolBitShapeDovetail default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("dovetail")
        self.assertEqual(shape["Diameter"].Value, 20.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["CuttingEdgeAngle"].Value, 60.0)
        self.assertEqual(unit(shape["CuttingEdgeAngle"]), "°")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("dovetail")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("CuttingEdgeAngle"), "Cutting angle")

    def test_toolbitshapeendmill_defaults(self):
        """Test ToolBitShapeEndmill default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("endmill")
        self.assertEqual(shape["Diameter"].Value, 5.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["CuttingEdgeHeight"].Value, 30.0)
        self.assertEqual(unit(shape["CuttingEdgeHeight"]), "mm")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("endmill")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("CuttingEdgeHeight"), "Cutting edge height")

    def test_toolbitshapeprobe_defaults(self):
        """Test ToolBitShapeProbe default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("probe")
        self.assertEqual(shape["Diameter"].Value, 6.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["ShaftDiameter"].Value, 4.0)
        self.assertEqual(unit(shape["ShaftDiameter"]), "mm")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("probe")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("Diameter"), "Ball diameter")
        self.assertEqual(instance.get_parameter_label("ShaftDiameter"), "Shaft diameter")

    def test_toolbitshapereamer_defaults(self):
        """Test ToolBitShapeReamer default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("reamer")
        self.assertEqual(shape["Diameter"].Value, 5.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["Length"].Value, 50.0)
        self.assertEqual(unit(shape["Length"]), "mm")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("reamer")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("Diameter"), "Diameter")

    def test_toolbitshapeslittingsaw_defaults(self):
        """Test ToolBitShapeSlittingSaw default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("slittingsaw")
        self.assertEqual(shape["Diameter"].Value, 100.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["BladeThickness"].Value, 3.0)
        self.assertEqual(unit(shape["BladeThickness"]), "mm")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("slittingsaw")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("BladeThickness"), "Blade thickness")

    def test_toolbitshapetap_defaults(self):
        """Test ToolBitShapeTap default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("tap")
        self.assertAlmostEqual(shape["Diameter"].Value, 8, 4)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["TipAngle"].Value, 90.0)
        self.assertEqual(unit(shape["TipAngle"]), "°")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("tap")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("TipAngle"), "Tip angle")

    def test_toolbitshapethreadmill_defaults(self):
        """Test ToolBitShapeThreadMill default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("thread-mill")
        self.assertEqual(shape["Diameter"].Value, 5.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["cuttingAngle"].Value, 60.0)
        self.assertEqual(unit(shape["cuttingAngle"]), "°")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("thread-mill")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("cuttingAngle"), "Cutting angle")

    def test_toolbitshapebullnose_defaults(self):
        """Test ToolBitShapeBullnose default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("bullnose")
        self.assertEqual(shape["Diameter"].Value, 5.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["CornerRadius"].Value, 1.5)
        self.assertEqual(unit(shape["CornerRadius"]), "mm")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("bullnose")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("CornerRadius"), "Corner radius")

    def test_toolbitshapevbit_defaults(self):
        """Test ToolBitShapeVBit default parameters and labels."""
        # Provide a dummy filepath for instantiation.
        shape = self._test_shape_common("v-bit")
        self.assertEqual(shape["Diameter"].Value, 10.0)
        self.assertEqual(unit(shape["Diameter"]), "mm")
        self.assertEqual(shape["CuttingEdgeAngle"].Value, 90.0)
        self.assertEqual(unit(shape["CuttingEdgeAngle"]), "°")
        self.assertEqual(shape["TipDiameter"].Value, 1.0)
        self.assertEqual(unit(shape["TipDiameter"]), "mm")
        # Need an instance to get parameter labels, get it from the asset manager
        uri = ToolBitShape.resolve_name("v-bit")
        instance = self.assets.get(uri)
        self.assertEqual(instance.get_parameter_label("CuttingEdgeAngle"), "Cutting edge angle")

    def test_serialize_deserialize(self):
        """
        Tests serialization and deserialization of a ToolBitShape object
        using the Asset interface methods.
        """
        # Load a shape instance from a fixture file
        fixture_path = (
            Path(__file__).parent / "Tools" / "Shape" / "test-path-tool-bit-shape-00.fcstd"
        )
        original_shape = ToolBitShape.from_file(fixture_path)

        # Serialize the shape using the to_bytes method
        serialized_data = original_shape.to_bytes(DummyAssetSerializer)

        # Assert that the serialized data is bytes and not empty
        self.assertIsInstance(serialized_data, bytes)
        self.assertTrue(len(serialized_data) > 0)

        # Deserialize the data using the from_bytes classmethod
        # Provide an empty dependencies mapping for this test
        deserialized_shape = ToolBitShape.from_bytes(
            serialized_data, original_shape.get_id(), {}, DummyAssetSerializer
        )

        # Assert that the deserialized object is a ToolBitShape instance
        self.assertIsInstance(deserialized_shape, ToolBitShape)
        # Assert that the deserialized shape has the same parameters as the original
        self.assertEqual(original_shape.get_parameters(), deserialized_shape.get_parameters())
        self.assertEqual(original_shape.name, deserialized_shape.name)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2021 FreeCAD Developers                                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the Layer object."""
## @package view_layer
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the Layer object.

## \addtogroup draftviewproviders
# @{
from pivy import coin
from PySide import QtCore
from PySide import QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
from draftobjects.layer import get_layer
from draftutils import params
from draftutils import utils
from draftutils.translate import translate


class ViewProviderLayer:
    """The viewprovider for the Layer object."""

    def __init__(self, vobj):
        self.Object = vobj.Object
        self.set_properties(vobj)

        vobj.Proxy = self

    def set_properties(self, vobj):
        """Set the properties only if they don't already exist."""
        properties = vobj.PropertiesList
        self.set_override_options(vobj, properties)
        self.set_visual_properties(vobj, properties)

    def set_override_options(self, vobj, properties):
        """Set property options only if they don't already exist."""
        if "OverrideLineColorChildren" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "If it is true, the objects contained "
                "within this layer will adopt "
                "the line color of the layer",
            )
            vobj.addProperty(
                "App::PropertyBool", "OverrideLineColorChildren", "Layer", _tip, locked=True
            )
            vobj.OverrideLineColorChildren = True

        if "OverrideShapeAppearanceChildren" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "If it is true, the objects contained "
                "within this layer will adopt "
                "the shape appearance of the layer",
            )
            vobj.addProperty(
                "App::PropertyBool", "OverrideShapeAppearanceChildren", "Layer", _tip, locked=True
            )
            vobj.OverrideShapeAppearanceChildren = True

        if "UsePrintColor" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "If it is true, the print color "
                "will be used when objects in this "
                "layer are placed on a TechDraw page",
            )
            vobj.addProperty("App::PropertyBool", "UsePrintColor", "Print", _tip, locked=True)

    def set_visual_properties(self, vobj, properties):
        """Set visual properties only if they don't already exist."""
        if "LineColor" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The line color of the objects " "contained within this layer"
            )
            vobj.addProperty("App::PropertyColor", "LineColor", "Layer", _tip, locked=True)
            vobj.LineColor = params.get_param_view("DefaultShapeLineColor") | 0x000000FF

        if "ShapeColor" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The shape color of the objects " "contained within this layer"
            )
            vobj.addProperty(
                "App::PropertyColor", "ShapeColor", "Layer", _tip, 4, locked=True
            )  # Hidden
            vobj.ShapeColor = params.get_param_view("DefaultShapeColor") | 0x000000FF

        if "ShapeAppearance" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The shape appearance of the objects " "contained within this layer",
            )
            vobj.addProperty(
                "App::PropertyMaterialList", "ShapeAppearance", "Layer", _tip, locked=True
            )
            material = App.Material()
            material.DiffuseColor = params.get_param_view("DefaultShapeColor") | 0x000000FF
            vobj.ShapeAppearance = (material,)

        if "LineWidth" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The line width of the objects contained " "within this layer"
            )
            vobj.addProperty("App::PropertyFloat", "LineWidth", "Layer", _tip, locked=True)
            vobj.LineWidth = params.get_param_view("DefaultShapeLineWidth")

        if "DrawStyle" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The draw style of the objects contained " "within this layer"
            )
            vobj.addProperty("App::PropertyEnumeration", "DrawStyle", "Layer", _tip, locked=True)
            vobj.DrawStyle = utils.DRAW_STYLES
            vobj.DrawStyle = params.get_param("DefaultDrawStyle")

        if "Transparency" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The transparency of the objects " "contained within this layer"
            )
            vobj.addProperty("App::PropertyPercent", "Transparency", "Layer", _tip, locked=True)
            vobj.Transparency = params.get_param_view("DefaultShapeTransparency")

        if "LinePrintColor" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The line color of the objects "
                "contained within this layer, "
                "when used on a TechDraw page",
            )
            vobj.addProperty("App::PropertyColor", "LinePrintColor", "Print", _tip, locked=True)
            vobj.LinePrintColor = params.get_param("DefaultPrintColor")

    def getIcon(self):
        """Return the path to the icon used by the viewprovider.

        Normally it returns the basic Layer icon, but if `icondata` exists
        it is the modified icon with the line and shape colors of the layer.
        """
        if hasattr(self, "icondata"):
            return self.icondata
        return ":/icons/Draft_Layer.svg"

    def attach(self, vobj):
        """Set up the scene sub-graph of the viewprovider."""
        self.Object = vobj.Object
        sep = coin.SoGroup()
        vobj.addDisplayMode(sep, "Default")

    def claimChildren(self):
        """Return objects that will be placed under it in the tree view.

        These are the elements of the `Group` property of the Proxy object.
        """
        if hasattr(self, "Object") and hasattr(self.Object, "Group"):
            if getattr(self.Object.ViewObject, "HideChildren", False):
                return []
            return self.Object.Group

    def getDisplayModes(self, vobj):
        """Return the display modes that this viewprovider supports."""
        return ["Default"]

    def getDefaultDisplayMode(self):
        """Return the default display mode."""
        return "Default"

    def setDisplayMode(self, mode):
        """Return the saved display mode."""
        return mode

    def dumps(self):
        """Return a tuple of objects to save or None."""
        return None

    def loads(self, state):
        """Set the internal properties from the restored state."""
        return None

    def updateData(self, obj, prop):
        """Execute when a property from the Proxy class is changed."""
        if prop == "Label":
            self._paint_tree_icon(obj.ViewObject)

    def change_view_properties(self, vobj, prop, old_prop=None, targets=None):
        """Change the properties of the targets, or of all objects in the layer group.

        A child's property is only changed if it is not overridden (if its value
        matches old_prop).
        """

        def _color_is_same(col1, col2):
            # Ignore alpha and round RGB values
            return [round(c, 2) for c in col1[:3]] == [round(c, 2) for c in col2[:3]]

        def _app_material_is_same(mat1, mat2):
            for prop in (
                "AmbientColor",
                "DiffuseColor",
                "EmissiveColor",
                "Shininess",
                "SpecularColor",
                "Transparency",
            ):
                if "Color" in prop:
                    if not _color_is_same(getattr(mat1, prop), getattr(mat2, prop)):
                        return False
                elif getattr(mat1, prop) != getattr(mat2, prop):
                    return False
            return True

        def _prop_is_same(prop1, prop2):
            if isinstance(prop1, tuple):
                if isinstance(prop1[0], App.Material):
                    # We do not check the length of the ShapeAppearance
                    return _app_material_is_same(prop1[0], prop2[0])
                return _color_is_same(prop1, prop2)
            return prop1 == prop2

        # Return if the property does not exist
        if not hasattr(vobj, prop):
            return
        # If the override properties are not set return without change
        if prop == "LineColor" and not vobj.OverrideLineColorChildren:
            return
        if prop == "ShapeAppearance" and not vobj.OverrideShapeAppearanceChildren:
            return

        for target_obj in (targets if targets is not None else vobj.Object.Group):
            target_vobj = target_obj.ViewObject

            if hasattr(target_vobj, prop):
                if old_prop is None or _prop_is_same(getattr(target_vobj, prop), old_prop):
                    setattr(target_vobj, prop, getattr(vobj, prop))

            # Use the line color for the point color and text color,
            # and the line width for the point size
            dic = {"LineColor": ("PointColor", "TextColor"), "LineWidth": ("PointSize",)}
            if prop in dic:
                for target_prop in dic[prop]:
                    if hasattr(target_vobj, target_prop):
                        if old_prop is None or _prop_is_same(
                            getattr(target_vobj, target_prop), old_prop
                        ):
                            setattr(target_vobj, target_prop, getattr(vobj, prop))

    def onBeforeChange(self, vobj, prop):
        if prop in ("LineColor", "ShapeAppearance", "LineWidth", "DrawStyle", "Visibility"):
            setattr(self, "old" + prop, getattr(vobj, prop))

    def onChanged(self, vobj, prop):
        """Execute when a view property is changed."""

        if not hasattr(vobj, prop):
            return

        if prop == "ShapeColor":
            if hasattr(vobj, "ShapeAppearance"):
                material = vobj.ShapeAppearance[0]
                if material.DiffuseColor != vobj.ShapeColor:
                    material.DiffuseColor = vobj.ShapeColor
                    vobj.ShapeAppearance = (material,)
            # The changed ShapeAppearance will do the rest:
            return

        if prop == "Transparency":
            if hasattr(vobj, "ShapeAppearance"):
                material = vobj.ShapeAppearance[0]
                if material.Transparency != vobj.Transparency / 100:
                    material.Transparency = vobj.Transparency / 100
                    vobj.ShapeAppearance = (material,)
            # The changed ShapeAppearance will do the rest:
            return

        if (
            prop == "ShapeAppearance"
            and hasattr(vobj, "ShapeColor")
            and hasattr(vobj, "Transparency")
        ):
            material = vobj.ShapeAppearance[0]
            if material.DiffuseColor != vobj.ShapeColor:
                vobj.ShapeColor = material.DiffuseColor
            if material.Transparency != vobj.Transparency / 100:
                vobj.Transparency = int(material.Transparency * 100)

        if (
            prop in ("LineColor", "ShapeAppearance", "LineWidth", "DrawStyle", "Visibility")
            and hasattr(vobj, "OverrideLineColorChildren")
            and hasattr(vobj, "OverrideShapeAppearanceChildren")
        ):
            old_prop = getattr(self, "old" + prop, None)
            self.change_view_properties(vobj, prop, old_prop)
            if hasattr(self, "old" + prop):
                delattr(self, "old" + prop)

        if prop in ("LineColor", "ShapeAppearance"):
            self._paint_tree_icon(vobj)

    def _paint_tree_icon(self, vobj):
        """Paint the layer icon in the tree view."""
        if not hasattr(vobj, "LineColor"):
            return
        if not hasattr(vobj, "ShapeAppearance"):
            return
        l_color = vobj.LineColor
        s_color = vobj.ShapeAppearance[0].DiffuseColor

        l_color = QtGui.QColor(int(l_color[0] * 255), int(l_color[1] * 255), int(l_color[2] * 255))
        s_color = QtGui.QColor(int(s_color[0] * 255), int(s_color[1] * 255), int(s_color[2] * 255))
        p1 = QtCore.QPointF(2, 17)
        p2 = QtCore.QPointF(13, 8)
        p3 = QtCore.QPointF(30, 15)
        p4 = QtCore.QPointF(20, 25)

        image = QtGui.QImage(32, 32, QtGui.QImage.Format_ARGB32)
        image.fill(QtCore.Qt.transparent)

        pt = QtGui.QPainter(image)
        pt.setBrush(QtGui.QBrush(s_color, QtCore.Qt.SolidPattern))
        pt.drawPolygon([p1, p2, p3, p4])
        pt.setPen(QtGui.QPen(l_color, 2, QtCore.Qt.SolidLine, QtCore.Qt.FlatCap))
        pt.drawPolygon([p1, p2, p3, p4])
        pt.end()

        byte_array = QtCore.QByteArray()
        buffer = QtCore.QBuffer(byte_array)
        buffer.open(QtCore.QIODevice.WriteOnly)
        image.save(buffer, "XPM")

        self.icondata = byte_array.data().decode("latin1")
        vobj.signalChangeIcon()

    def canDragObject(self, obj):
        """Return True to allow dragging one object from the Layer.

        Also store parent group data for update_groups_after_drag_drop and
        trigger that function.
        """
        if not hasattr(self, "old_parent_data"):
            self.old_parent_data = {}
        old_data = []
        for parent in obj.InList:
            if hasattr(parent, "Group"):
                old_data.append([parent, parent.Group])
        # Layers are not in the Inlist because a layer's Group is App::PropertyLinkListHidden:
        layer = get_layer(obj)
        if layer is not None:
            old_data.append([layer, layer.Group])
        if old_data:
            self.old_parent_data.setdefault(obj, old_data)
            QtCore.QTimer.singleShot(0, self.update_groups_after_drag_drop)

        return True

    def canDragObjects(self):
        """Return True to allow dragging many objects from the Layer."""
        return True

    def dragObject(self, vobj, child):
        """Remove the object that was dragged from the layer."""
        obj = vobj.Object
        obj.Proxy.removeObject(obj, child)
        App.ActiveDocument.recompute()

    def canDropObject(self, obj):
        """Return true to allow dropping one object.

        If the object being dropped is itself a `'Layer'`, return `False`
        to prevent dropping a layer inside a layer, at least for now.

        Also store parent group data for update_groups_after_drag_drop and
        trigger that function.
        """
        if utils.get_type(obj) == "Layer":
            return False
        return self.canDragObject(obj)

    def canDropObjects(self):
        """Return true to allow dropping many objects."""
        return True

    def dropObject(self, vobj, child):
        """Add the object that was dropped on the Layer to the group.

        This also results in a call to `change_view_properties` to update the
        view properties of the child.
        """
        obj = vobj.Object
        obj.Proxy.addObject(obj, child)
        App.ActiveDocument.recompute()

    def update_groups_after_drag_drop(self):
        """Workaround function to improve the drag and drop behavior of Layer
        objects.

        The function processes the parent group data stored in the
        old_parent_data dictionary by canDragObject and canDropObject.
        """

        # The function can be called multiple times, old_parent_data will be
        # empty after the first call.
        if (not hasattr(self, "old_parent_data")) or (not self.old_parent_data):
            return

        # List to collect parents whose Group must be updated.
        # This has to happen later in a separate loop as we need the unmodified
        # InList properties of the children in the main loop.
        parents_to_update = []

        # Main loop:
        for child, old_data in self.old_parent_data.items():

            # We assume a single old and a single new layer...

            old_layer = None
            for old_parent, old_parent_group in old_data:
                if utils.get_type(old_parent) == "Layer":
                    old_layer = old_parent
                    break

            new_layer = get_layer(child)
            if new_layer == old_layer:
                continue

            elif new_layer is None:
                # An object was dragged out of a layer.
                # We need to check if it was put in a new group. If that is
                # the case the content of old_layer should be restored.
                # If the object was not put in a new group it was dropped on
                # the document node, in that case we do nothing.
                old_parents = [sub[0] for sub in old_data]
                for new_parent in child.InList:
                    if (
                        hasattr(new_parent, "Group") and new_parent not in old_parents
                    ):  # New group check.
                        for old_parent, old_parent_group in old_data:
                            if old_parent == old_layer:
                                parents_to_update.append([old_parent, old_parent_group])
                                break
                        break

            else:
                # A new layer was assigned.
                # The content of all `non-layer` groups should be restored.
                for old_parent, old_parent_group in old_data:
                    if utils.get_type(old_parent) != "Layer":
                        parents_to_update.append([old_parent, old_parent_group])

        # Update parents:
        if parents_to_update:
            for old_parent, old_parent_group in parents_to_update:
                old_parent.Group = old_parent_group
            App.ActiveDocument.recompute()

        self.old_parent_data = {}

    def replaceObject(self, old_obj, new_obj):
        """Return immediately to prevent replacement of children."""
        return

    def setupContextMenu(self, vobj, menu):
        """Set up actions to perform in the context menu."""
        action_activate = QtGui.QAction(
            QtGui.QIcon(":/icons/button_right.svg"), translate("draft", "Activate Layer"), menu
        )
        action_activate.triggered.connect(self.activate)
        menu.addAction(action_activate)

        action_reassign = QtGui.QAction(
            QtGui.QIcon(":/icons/Draft_Apply.svg"),
            translate("draft", "Reassign Properties of Layer"),
            menu,
        )
        action_reassign.triggered.connect(self.reassign_props)
        menu.addAction(action_reassign)

        action_select = QtGui.QAction(
            QtGui.QIcon(":/icons/Draft_SelectGroup.svg"),
            translate("draft", "Select Layer Contents"),
            menu,
        )
        action_select.triggered.connect(self.select_contents)
        menu.addAction(action_select)

    def activate(self):
        """Activate the selected layer, it becomes the Autogroup."""
        Gui.Selection.clearSelection()
        Gui.Selection.addSelection(self.Object)
        if not "Draft_AutoGroup" in Gui.listCommands():
            Gui.activateWorkbench("DraftWorkbench")
        Gui.runCommand("Draft_AutoGroup")

    def reassign_props(self):
        for prop in ("LineColor", "ShapeAppearance", "LineWidth", "DrawStyle", "Visibility"):
            self.onChanged(self.Object.ViewObject, prop)

    def select_contents(self):
        """Select the contents of the layer."""
        Gui.Selection.clearSelection()
        for obj in self.Object.Group:
            Gui.Selection.addSelection(obj)


class ViewProviderLayerContainer:
    """The viewprovider for the LayerContainer object."""

    def __init__(self, vobj):
        self.Object = vobj.Object
        vobj.Proxy = self

    def getIcon(self):
        """Return the path to the icon used by the viewprovider."""
        return ":/icons/Draft_Layers.svg"

    def attach(self, vobj):
        """Set up the scene sub-graph of the viewprovider."""
        self.Object = vobj.Object

    def setupContextMenu(self, vobj, menu):
        """Set up actions to perform in the context menu."""
        action_add = QtGui.QAction(
            QtGui.QIcon(":/icons/Draft_NewLayer.svg"), translate("draft", "Add New Layer"), menu
        )
        action_add.triggered.connect(self.add_layer)
        menu.addAction(action_add)

        action_reassign = QtGui.QAction(
            QtGui.QIcon(":/icons/Draft_Apply.svg"),
            translate("draft", "Reassign Properties of All Layers"),
            menu,
        )
        action_reassign.triggered.connect(self.reassign_props)
        menu.addAction(action_reassign)

        action_merge = QtGui.QAction(
            QtGui.QIcon(":/icons/Draft_Layers.svg"),
            translate("draft", "Merge Layer Duplicates"),
            menu,
        )
        action_merge.triggered.connect(self.merge_by_name)
        menu.addAction(action_merge)

    def add_layer(self):
        """Creates a new layer"""
        import Draft

        doc = App.ActiveDocument
        doc.openTransaction(translate("draft", "Add New Layer"))

        Draft.make_layer(
            name=None,
            line_color=None,
            shape_color=None,
            line_width=None,
            draw_style=None,
            transparency=None,
        )

        doc.recompute()
        doc.commitTransaction()

    def reassign_props(self):
        for obj in self.Object.Group:
            if utils.get_type(obj) == "Layer":
                obj.ViewObject.Proxy.reassign_props()

    def merge_by_name(self):
        """Merge the layers that have the same base label."""
        doc = App.ActiveDocument
        doc.openTransaction(translate("draft", "Merge Layer Duplicates"))

        layer_container = self.Object
        layers = []
        for obj in layer_container.Group:
            if utils.get_type(obj) == "Layer":
                layers.append(obj)

        to_delete = []
        for layer in layers:
            # Remove trailing digits (usually 3 but there might be more) and
            # trailing spaces from Label before comparing:
            base_label = layer.Label.rstrip("0123456789 ")

            # Try to find the `'base'` layer:
            base = None
            for other_layer in layers:
                if (
                    (not other_layer in to_delete)  # Required if there are duplicate labels.
                    and other_layer != layer
                    and other_layer.Label.upper() == base_label.upper()
                ):
                    base = other_layer
                    break

            if base:
                if layer.Group:
                    base_group = base.Group
                    for obj in layer.Group:
                        if not obj in base_group:
                            base_group.append(obj)
                    base.Group = base_group
                to_delete.append(layer)
            elif layer.Label != base_label:
                layer.Label = base_label

        for layer in to_delete:
            doc.removeObject(layer.Name)

        doc.recompute()
        doc.commitTransaction()

    def dumps(self):
        """Return a tuple of objects to save or None."""
        return None

    def loads(self, state):
        """Set the internal properties from the restored state."""
        return None

    def replaceObject(self, old_obj, new_obj):
        """Return immediately to prevent replacement of children."""
        return


# Alias for compatibility with v0.18 and earlier
_ViewProviderVisGroup = ViewProviderLayer

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to activate the different snapping methods."""
## @package gui_snaps
# \ingroup draftguitools
# \brief Provides GUI tools to activate the different snapping methods.

## \addtogroup draftguitools
# @{
from PySide import QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import draftguitools.gui_base as gui_base

from draftutils.messages import _log
from draftutils.translate import translate


class Draft_Snap_Base:
    """Base Class inherited by all Draft Snap commands."""

    def Activated(self, status=0):
        # _log("GuiCommand: {}".format(self.__class__.__name__))

        if hasattr(Gui, "Snapper"):
            Gui.Snapper.toggle_snap(self.__class__.__name__[11:], bool(status))

    def IsActive(self):
        return hasattr(Gui, "Snapper") and Gui.Snapper.isEnabled("Lock")

    def isChecked(self):
        """Return true if the given snap is active in Snapper."""
        return hasattr(Gui, "Snapper") and self.__class__.__name__[11:] in Gui.Snapper.active_snaps


class Draft_Snap_Lock(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Lock tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Lock",
            "Accel": "Shift+S",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Lock", "Snap Lock"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Lock", "Enables or disables snapping globally"
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }

    def IsActive(self):
        return True


Gui.addCommand("Draft_Snap_Lock", Draft_Snap_Lock())


class Draft_Snap_Midpoint(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Midpoint tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Midpoint",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Midpoint", "Snap Midpoint"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Snap_Midpoint", "Snaps to the midpoint of edges"),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Midpoint", Draft_Snap_Midpoint())


class Draft_Snap_Perpendicular(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Perpendicular tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Perpendicular",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Perpendicular", "Snap Perpendicular"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Perpendicular", "Snaps to the perpendicular points on faces and edges"
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Perpendicular", Draft_Snap_Perpendicular())


class Draft_Snap_Grid(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Grid tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Grid",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Grid", "Snap Grid"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Grid", "Snaps to the intersections of grid lines"
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Grid", Draft_Snap_Grid())


class Draft_Snap_Intersection(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Intersection tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Intersection",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Intersection", "Snap Intersection"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Intersection",
                "Snaps to the intersection of 2 edges, and the intersection of a face and an edge",
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Intersection", Draft_Snap_Intersection())


class Draft_Snap_Parallel(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Parallel tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Parallel",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Parallel", "Snap Parallel"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Parallel", "Snaps to an imaginary line parallel to straight edges"
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Parallel", Draft_Snap_Parallel())


class Draft_Snap_Endpoint(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Endpoint tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Endpoint",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Endpoint", "Snap Endpoint"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Snap_Endpoint", "Snaps to the endpoints of edges"),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Endpoint", Draft_Snap_Endpoint())


class Draft_Snap_Angle(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Angle tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Angle",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Angle", "Snap Angle"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Angle",
                "Snaps to the special cardinal points on circular edges, at multiples of 30° and 45°",
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Angle", Draft_Snap_Angle())


class Draft_Snap_Center(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Center tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Center",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Center", "Snap Center"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Center",
                "Snaps to the center point of faces and circular edges, and to the placement point of working plane proxies and building parts",
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Center", Draft_Snap_Center())


class Draft_Snap_Extension(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Extension tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Extension",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Extension", "Snap Extension"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Extension",
                "Snaps to an imaginary line that extends beyond the endpoints of straight edges",
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Extension", Draft_Snap_Extension())


class Draft_Snap_Near(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Near tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Near",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Near", "Snap Near"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Near", "Snaps to the nearest point on faces and edges"
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Near", Draft_Snap_Near())


class Draft_Snap_Ortho(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Ortho tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Ortho",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Ortho", "Snap Ortho"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Ortho",
                "Snaps to imaginary lines that cross the previous point at multiples of 45°",
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Ortho", Draft_Snap_Ortho())


class Draft_Snap_Special(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Special tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Special",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Special", "Snap Special"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Special", "Snaps to special points defined by the object"
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Special", Draft_Snap_Special())


class Draft_Snap_Dimensions(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_Dimensions tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_Dimensions",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_Dimensions", "Snap Dimensions"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_Dimensions", "Shows temporary X and Y dimensions"
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_Dimensions", Draft_Snap_Dimensions())


class Draft_Snap_WorkingPlane(Draft_Snap_Base):
    """GuiCommand for the Draft_Snap_WorkingPlane tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap_WorkingPlane",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Snap_WorkingPlane", "Snap Working Plane"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Snap_WorkingPlane", "Projects snap points onto the current working plane"
            ),
            "CmdType": "NoTransaction",
            "Checkable": self.isChecked(),
        }


Gui.addCommand("Draft_Snap_WorkingPlane", Draft_Snap_WorkingPlane())


class ShowSnapBar(Draft_Snap_Base):
    """GuiCommand for the Draft_ShowSnapBar tool."""

    def GetResources(self):
        return {
            "Pixmap": "Draft_Snap",
            "MenuText": QT_TRANSLATE_NOOP("Draft_ShowSnapBar", "Show Snap Toolbar"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_ShowSnapBar", "Shows the snap toolbar if it is hidden"
            ),
            "CmdType": "NoTransaction",
        }

    def Activated(self):
        """Execute when the command is called."""
        if hasattr(Gui, "Snapper"):
            toolbar = Gui.Snapper.get_snap_toolbar()
            if toolbar is not None:
                toolbar.show()


Gui.addCommand("Draft_ShowSnapBar", ShowSnapBar())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "Netgen preference page class"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

from PySide import QtCore
from PySide import QtGui

import FreeCAD
import FreeCADGui


class DlgSettingsNetgen(QtGui.QWidget):

    def __init__(self):
        self.form = FreeCADGui.PySideUic.loadUi(":ui/DlgSettingsNetgen.ui")
        self.grp = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/Netgen")

        QtCore.QObject.connect(
            self.form.fc_netgen_python_path,
            QtCore.SIGNAL("fileNameSelected(QString)"),
            self.netgen_python_path_selected,
        )

    def loadSettings(self):
        self.form.ckb_legacy.setChecked(self.grp.GetBool("UseLegacyNetgen", True))
        self.form.sb_threads.setValue(
            self.grp.GetInt("NumOfThreads", QtCore.QThread.idealThreadCount())
        )
        self.form.fc_netgen_python_path.setProperty(
            "fileName", self.grp.GetString("NetgenPythonPath", "")
        )

        self.populate_log_verbosity()

    def saveSettings(self):
        self.grp.SetBool("UseLegacyNetgen", self.form.ckb_legacy.isChecked())
        self.grp.SetInt("LogVerbosity", self.form.cb_log_verbosity.currentData())
        self.grp.SetInt("NumOfThreads", self.form.sb_threads.value())
        self.grp.SetString("NetgenPythonPath", self.form.fc_netgen_python_path.property("fileName"))

    def populate_log_verbosity(self):
        values = {
            "None": 0,
            "Least": 1,
            "Little": 2,
            "Moderate": 3,
            "Much": 4,
            "Most": 5,
        }

        for v in values:
            self.form.cb_log_verbosity.addItem(v, values[v])

        current = self.grp.GetInt("LogVerbosity", 2)
        index = self.form.cb_log_verbosity.findData(current)
        self.form.cb_log_verbosity.setCurrentIndex(index)

    def netgen_python_path_selected(self, path):
        if path and not QtCore.QStandardPaths.findExecutable(path):
            QtGui.QMessageBox.critical(
                None, "Netgen", self.tr("Executable '{}' not found").format(path)
            )
